@@ -7,7 +7,7 @@ use IO::File;
 use Scalar::Util qw/weaken/;
 use Encode qw/decode/;
 
-our $VERSION = '0.990103';
+our $VERSION = '0.990104';
 
 use constant CHUNK_SIZE => 4096;
 
@@ -59,7 +59,6 @@ sub new {
 
     # Default attributes
     my $attrs = {};
-    $attrs->{pretty}       = 1;
     $attrs->{vars_as_subs} = 0;
     $attrs->{tape}         = [];
     $attrs->{encoding}     = 'utf-8';
@@ -107,9 +106,7 @@ sub vars_as_subs {
     @_ > 1 ? $_[0]->{vars_as_subs} = $_[1] : $_[0]->{vars_as_subs};
 }
 
-sub pretty   { @_ > 1 ? $_[0]->{pretty}   = $_[1] : $_[0]->{pretty} }
 sub format   { @_ > 1 ? $_[0]->{format}   = $_[1] : $_[0]->{format} }
-sub tape     { @_ > 1 ? $_[0]->{tape}     = $_[1] : $_[0]->{tape} }
 sub encoding { @_ > 1 ? $_[0]->{encoding} = $_[1] : $_[0]->{encoding} }
 
 sub escape_html {
@@ -124,6 +121,7 @@ sub filters  { @_ > 1 ? $_[0]->{filters}  = $_[1] : $_[0]->{filters} }
 sub prepend  { @_ > 1 ? $_[0]->{prepend}  = $_[1] : $_[0]->{prepend} }
 sub append   { @_ > 1 ? $_[0]->{append}   = $_[1] : $_[0]->{append} }
 sub escape   { @_ > 1 ? $_[0]->{escape}   = $_[1] : $_[0]->{escape} }
+sub tape     { @_ > 1 ? $_[0]->{tape}     = $_[1] : $_[0]->{tape} }
 sub vars     { @_ > 1 ? $_[0]->{vars}     = $_[1] : $_[0]->{vars} }
 
 sub helpers_arg {
@@ -253,7 +251,7 @@ sub parse {
         }
 
         # Filter
-        if ($line =~ m/^:(.*)/) {
+        if ($line =~ m/^:(\w+)/) {
             $el->{type} = 'filter';
             $el->{name} = $1;
             $el->{text} = '';
@@ -548,9 +546,11 @@ EOF
 
     my $stack = [];
 
+    my $output = '';
     my @lines;
     my $count    = 0;
     my $in_block = 0;
+  ELEM:
     for my $el (@{$self->tape}) {
         my $level = $el->{level};
         $level -= 2 * $in_block if $in_block;
@@ -558,20 +558,6 @@ EOF
         my $offset = '';
         $offset .= ' ' x $level;
 
-        my $prev_el = $self->tape->[$count - 1];
-        my $next_el = $self->tape->[$count + 1];
-
-        my $prev_stack_el = $stack->[-1];
-
-        if ($prev_stack_el && $prev_stack_el->{type} eq 'comment') {
-            if ($prev_stack_el->{level} == $el->{level}) {
-                pop @$stack;
-            }
-            else {
-                next;
-            }
-        }
-
         my $escape = '';
         if (   (!exists $el->{escape} && $self->escape_html)
             || (exists $el->{escape} && $el->{escape} == 1))
@@ -579,10 +565,27 @@ EOF
             $escape = 'escape';
         }
 
+        my $prev_el = $self->tape->[$count - 1];
+        my $next_el = $self->tape->[$count + 1];
+
+        my $prev_stack_el = $stack->[-1];
+
+        if ($prev_stack_el && $prev_stack_el->{type} eq 'comment') {
+            if (   $el->{line}
+                && $prev_stack_el->{level} >= $el->{level})
+            {
+                pop @$stack;
+            }
+            else {
+                next ELEM;
+            }
+        }
+
         if (   $el->{line}
             && $prev_stack_el
             && $prev_stack_el->{level} >= $el->{level})
         {
+          STACKEDBLK:
             while (my $poped = pop @$stack) {
                 my $level = $poped->{level};
                 $level -= 2 * $in_block if $in_block;
@@ -601,182 +604,208 @@ EOF
                     push @lines, qq|\$_H .= "$poped_offset$ending\n";|;
                 }
 
-                last if $poped->{level} == $el->{level};
+                last STACKEDBLK if $poped->{level} == $el->{level};
             }
         }
 
-        my $output = '';
-        if ($el->{type} eq 'tag') {
-            my $ending =
-              $el->{autoclose} && $self->format eq 'xhtml' ? ' /' : '';
 
-            my $attrs = '';
-            if ($el->{attrs}) {
-                for (my $i = 0; $i < @{$el->{attrs}}; $i += 2) {
-                    my $name  = $el->{attrs}->[$i];
-                    my $value = $el->{attrs}->[$i + 1];
-                    my $text  = $value->{text};
+      SWITCH: {
 
-                    if ($name eq 'class') {
-                        $el->{class} ||= [];
-                        if ($value->{type} eq 'text') {
-                            push @{$el->{class}}, $text;
-                        }
-                        else {
-                            push @{$el->{class}}, qq/" . $text . "/;
-                        }
-                        next;
-                    }
-                    elsif ($name eq 'id') {
-                        $el->{id} ||= '';
-                        $el->{id} = $el->{id} . '_' if $el->{id};
-                        $el->{id} .= $value->{text};
-                        next;
-                    }
+            if ($el->{type} eq 'tag') {
+                my $ending =
+                  $el->{autoclose} && $self->format eq 'xhtml' ? ' /' : '';
 
-                    if ($value->{type} eq 'text' || $value->{type} eq 'expr')
-                    {
-                        $attrs .= ' ';
-                        $attrs .= $name;
-                        $attrs .= '=';
+                my $attrs = '';
+                if ($el->{attrs}) {
+                  ATTR:
+                    for (my $i = 0; $i < @{$el->{attrs}}; $i += 2) {
+                        my $name  = $el->{attrs}->[$i];
+                        my $value = $el->{attrs}->[$i + 1];
+                        my $text  = $value->{text};
 
-                        if ($value->{type} eq 'text') {
-                            $attrs .= "'" . $self->_parse_text($text) . "'";
+                        if ($name eq 'class') {
+                            $el->{class} ||= [];
+                            if ($value->{type} eq 'text') {
+                                push @{$el->{class}}, $text;
+                            }
+                            else {
+                                push @{$el->{class}}, qq/" . $text . "/;
+                            }
+                            next ATTR;
                         }
-                        else {
-                            $attrs .= qq/'" . $text . "'/;
+                        elsif ($name eq 'id') {
+                            $el->{id} ||= '';
+                            $el->{id} = $el->{id} . '_' if $el->{id};
+                            $el->{id} .= $value->{text};
+                            next ATTR;
                         }
-                    }
-                    elsif ($value->{type} eq 'boolean' && $value->{text}) {
-                        $attrs .= ' ';
-                        $attrs .= $name;
-                        if ($self->format eq 'xhtml') {
+
+                        if (   $value->{type} eq 'text'
+                            || $value->{type} eq 'expr')
+                        {
+                            $attrs .= ' ';
+                            $attrs .= $name;
                             $attrs .= '=';
-                            $attrs .= qq/'$name'/;
+
+                            if ($value->{type} eq 'text') {
+                                $attrs
+                                  .= "'" . $self->_parse_text($text) . "'";
+                            }
+                            else {
+                                $attrs .= qq/'" . $text . "'/;
+                            }
                         }
-                    }
+                        elsif ($value->{type} eq 'boolean' && $value->{text})
+                        {
+                            $attrs .= ' ';
+                            $attrs .= $name;
+                            if ($self->format eq 'xhtml') {
+                                $attrs .= '=';
+                                $attrs .= qq/'$name'/;
+                            }
+                        }
+                    }    #end:for ATTR
                 }
+
+                my $tail = '';
+                if ($el->{class}) {
+                    $tail .= qq/ class='"./;
+                    $tail .= qq/join(' ', sort(/;
+                    $tail .= join(',', map {"\"$_\""} @{$el->{class}});
+                    $tail .= qq/))/;
+                    $tail .= qq/."'/;
+                }
+
+                if ($el->{id}) {
+                    $tail .= qq/ id='$el->{id}'/;
+                }
+
+                $output .= qq|"$offset<$el->{name}$tail$attrs$ending>"|;
+
+                if ($el->{text} && $el->{expr}) {
+                    $output .= '. (do {' . $el->{text} . '} || "")';
+                    $output .= qq| . "</$el->{name}>"|;
+                }
+                elsif ($el->{text}) {
+                    $output .= qq/. $escape / . '"'
+                      . $self->_parse_text($el->{text}) . '";';
+                    $output .= qq|\$_H .= "</$el->{name}>"|
+                      unless $el->{autoclose};
+                }
+                elsif (
+                    !$next_el
+                    || (   $next_el
+                        && $next_el->{level} <= $el->{level})
+                  )
+                {
+                    $output .= qq|. "</$el->{name}>"| unless $el->{autoclose};
+                }
+                elsif (!$el->{autoclose}) {
+                    push @$stack, $el;
+                }
+
+                $output .= qq|. "\n"|;
+                $output .= qq|;|;
+                last SWITCH;
             }
 
-            my $tail = '';
-            if ($el->{class}) {
-                $tail .= qq/ class='"./;
-                $tail .= qq/join(' ', sort(/;
-                $tail .= join(',', map {"\"$_\""} @{$el->{class}});
-                $tail .= qq/))/;
-                $tail .= qq/."'/;
+            if ($el->{line} && $el->{type} eq 'text') {
+                $output = qq/"$offset"/;
+
+                $el->{text} = '' unless defined $el->{text};
+
+                if ($el->{expr}) {
+                    $output .= qq/. $escape / . +$el->{text};
+                    $output .= qq/;\$_H .= "\n"/;
+                }
+                elsif ($el->{text}) {
+                    $output
+                      .= '.'
+                      . qq/$escape / . '"'
+                      . $self->_parse_text($el->{text}) . '"';
+                    $output .= qq/. "\n"/;
+                }
+
+                $output .= qq/;/;
+                last SWITCH;
             }
 
-            if ($el->{id}) {
-                $tail .= qq/ id='$el->{id}'/;
-            }
-
-            $output .= qq|"$offset<$el->{name}$tail$attrs$ending>"|;
-
-            if ($el->{text} && $el->{expr}) {
-                $output .= '. (do {' . $el->{text} . '} || "")';
-                $output .= qq| . "</$el->{name}>"|;
-            }
-            elsif ($el->{text}) {
-                $output .= qq/. $escape / . '"'
-                  . $self->_parse_text($el->{text}) . '";';
-                $output .= qq|\$_H .= "</$el->{name}>"|
-                  unless $el->{autoclose};
-            }
-            elsif (
-                !$next_el
-                || (   $next_el
-                    && $next_el->{level} <= $el->{level})
-              )
-            {
-                $output .= qq|. "</$el->{name}>"| unless $el->{autoclose};
-            }
-            elsif (!$el->{autoclose}) {
+            if ($el->{type} eq 'block') {
+                push @lines,  $el->{text};
                 push @$stack, $el;
+
+                if ($prev_el && $prev_el->{level} > $el->{level}) {
+                    $in_block--;
+                }
+
+                if ($next_el && $next_el->{level} > $el->{level}) {
+                    $in_block++;
+                }
+                last SWITCH;
             }
 
-            $output .= qq|. "\n"|;
-            $output .= qq|;|;
-        }
-        elsif ($el->{type} eq 'text') {
-            $output = qq/"$offset"/;
+            if ($el->{type} eq 'html_comment') {
+                $output = qq/"$offset"/;
 
-            $el->{text} = '' unless defined $el->{text};
+                $output .= qq/ . "<!--"/;
+                $output .= qq/ . "[if $el->{if}]>"/ if $el->{if};
 
-            if ($el->{expr}) {
-                $output .= qq/. $escape / . +$el->{text};
-                $output .= qq/;\$_H .= "\n"/;
-            }
-            elsif ($el->{text}) {
-                $output
-                  .= '.'
-                  . qq/$escape / . '"'
-                  . $self->_parse_text($el->{text}) . '"';
-                $output .= qq/. "\n"/;
+                if ($el->{text}) {
+                    $output .= qq/. " $el->{text} -->\n"/;
+                }
+                else {
+                    $output .= qq/. "\n"/;
+                    push @$stack, $el;
+                }
+
+                $output .= qq/;/;
+                last SWITCH;
             }
 
-            $output .= qq/;/;
-        }
-        elsif ($el->{type} eq 'block') {
-            push @lines,  $el->{text};
-            push @$stack, $el;
-
-            if ($prev_el && $prev_el->{level} > $el->{level}) {
-                $in_block--;
-            }
-
-            if ($next_el && $next_el->{level} > $el->{level}) {
-                $in_block++;
-            }
-        }
-        elsif ($el->{type} eq 'html_comment') {
-            $output = qq/"$offset"/;
-
-            $output .= qq/ . "<!--"/;
-            $output .= qq/ . "[if $el->{if}]>"/ if $el->{if};
-
-            if ($el->{text}) {
-                $output .= qq/. " $el->{text} -->\n"/;
-            }
-            else {
-                $output .= qq/. "\n"/;
+            if ($el->{type} eq 'comment') {
                 push @$stack, $el;
+                last SWITCH;
             }
 
-            $output .= qq/;/;
-        }
-        elsif ($el->{type} eq 'comment') {
-            push @$stack, $el;
-        }
-        elsif ($el->{type} eq 'filter') {
-            my $filter = $self->filters->{$el->{name}};
-            die "unknown filter: $el->{name}" unless $filter;
+            if ($el->{type} eq 'filter') {
+                my $filter = $self->filters->{$el->{name}};
+                die "unknown filter: $el->{name}" unless $filter;
 
-            if ($el->{name} eq 'escaped') {
-                $output =
-                  qq/escape "/ . $self->_parse_text($el->{text}) . qq/\n";/;
-            }
-            else {
-                $el->{text} = $filter->($el->{text});
+                if ($el->{name} eq 'escaped') {
+                    $output =
+                        qq/escape "/
+                      . $self->_parse_text($el->{text})
+                      . qq/\n";/;
+                }
+                else {
+                    $el->{text} = $filter->($el->{text});
 
-                my $text = $self->_parse_text($el->{text});
-                $text =~ s/\\\n/\\n/g;
-                $output = qq/"/ . $text . qq/\n";/;
+                    my $text = $self->_parse_text($el->{text});
+                    $text =~ s/\\\n/\\n/g;
+                    $output = qq/"/ . $text . qq/\n";/;
+                }
+                last SWITCH;
             }
-        }
-        else {
+
+            unless ($el->{text}) {
+                last SWITCH;
+            }
+
             die "unknown type=" . $el->{type};
-        }
 
+        }    #end:SWITCH
+    }    #end:ELEM
+    continue {
         push @lines, '$_H .= ' . $output if $output;
-
+        $output = '';
         $count++;
-    }
+    }    #ELEM
 
     my $last_empty_line = 0;
     $last_empty_line = 1
       if $self->tape->[-1] && $self->tape->[-1]->{line} eq '';
 
+    # Close remaining content blocks, last-seen first
     foreach my $el (reverse @$stack) {
         my $offset = ' ' x $el->{level};
         my $ending = '';
@@ -1027,21 +1056,29 @@ features. Do not expect Ruby specific things to work.
 
 L<Text::Haml> implements the following attributes:
 
-=head2 C<format>
+=head2 C<append>
 
-    Supported formats: xhtml, html, html5.
+Holds the string of code that is appended to the generated Perl code.
 
-    Default is xhtml.
+=head2 C<code>
+
+Holds the Perl code.
+
+=head2 C<compiled>
+
+Holds compiled code.
 
 =head2 C<encoding>
 
-    Default is utf-8.
+    $haml->encoding('utf-8');
+
+Default is utf-8.
 
 =head2 C<escape>
 
-    Escape subroutine presented as string.
+Escape subroutine presented as string.
 
-    Default is
+Default is
 
     $haml->escape(<<'EOF');
         my $s = shift;
@@ -1055,14 +1092,38 @@ L<Text::Haml> implements the following attributes:
 
 =head2 C<escape_html>
 
-    Switch on/off Haml output html escaping.
+    $haml->escape_html(0);
 
-    Default is on.
+Switch on/off Haml output html escaping. Default is on.
+
+=head2 C<filters>
+
+Holds filters.
+
+=head2 C<format>
+
+    $haml->format('xhtml');
+
+Supported formats: xhtml, html, html5.
+
+Default is xhtml.
+
+=head2 C<namespace>
+
+Holds the namespace under which the Perl package is generated.
+
+=head2 C<prepend>
+
+Holds the string of code that is prepended to the generated Perl code.
+
+=head2 C<vars>
+
+Holds the variables that are passed during the rendering.
 
 =head2 C<vars_as_subs>
 
 When options is B<NOT SET> (by default) passed variables are normal Perl
-variables and are use with C<$> prefix.
+variables and are used with C<$> prefix.
 
     $haml->render('%p $var', var => 'hello');
 
@@ -1081,9 +1142,6 @@ prefix.
 
 =head2 C<helpers>
 
-    Holds helpers subroutines. Helpers can be called in Haml text as normal Perl
-    functions. See also add_helper.
-
     helpers => {
         foo => sub {
             my $self   = shift;
@@ -1095,17 +1153,24 @@ prefix.
         }
     }
 
-=head2 C<helpers_arg>
+Holds helpers subroutines. Helpers can be called in Haml text as normal Perl
+functions. See also add_helper.
 
-    First argument passed to the helper.
+=head2 C<helpers_arg>
 
     $haml->helpers_args($my_context);
 
-    Deafault is Text::Haml instance.
+First argument passed to the helper (L<Text::Haml> instance by default).
 
 =head2 C<error>
 
-    Holds last error.
+    $haml->error;
+
+Holds the last error.
+
+=head2 C<tape>
+
+Holds parsed haml elements.
 
 =head1 METHODS
 
@@ -1115,30 +1180,54 @@ prefix.
 
 =head2 C<add_helper>
 
-    Adds a new helper.
-
     $haml->add_helper(current_time => sub { time });
 
+Adds a new helper.
+
 =head2 C<add_filter>
 
-    Adds a new filter.
-
     $haml->add_filter(compress => sub { $_[0] =~ s/\s+/ /g; $_[0]});
 
-=head2 C<render>
+Adds a new filter.
 
-    Renders Haml string. Returns undef on error. See error attribute.
+=head2 C<build>
+
+    $haml->build(@_);
+
+Builds the Perl code.
+
+=head2 C<compile>
+
+    $haml->compile;
+
+Compiles parsed code.
+
+=head2 C<interpret>
+
+    $haml->interpret(@_);
+
+Interprets compiled code.
+
+=head2 C<parse>
+
+    $haml->parse('%p foo');
+
+Parses Haml string building a tree.
+
+=head2 C<render>
 
     my $text = $haml->render('%p foo');
 
     my $text = $haml->render('%p var', var => 'hello');
 
+Renders Haml string. Returns undef on error. See error attribute.
+
 =head2 C<render_file>
 
-    A helper method that loads a file and passes it to the render method.
-
     my $text = $haml->render_file('foo.haml');
 
+A helper method that loads a file and passes it to the render method.
+
 =head1 PERL SPECIFIC IMPLEMENTATION ISSUES
 
 =head2 String interpolation
@@ -1163,7 +1252,7 @@ $haml->render("%a{href => 'bar'}");
 
 =head2 Repository
 
-    http://github.com/vti/text-haml/commits/master
+    http://github.com/vti/text-haml
 
 =head1 AUTHOR
 
@@ -1173,11 +1262,13 @@ Viacheslav Tykhanovskyi, C<vti@cpan.org>.
 
 In alphabetical order:
 
+Nick Ragouzis
+
 Norman Clarke
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2009, Viacheslav Tykhanovskyi.
+Copyright (C) 2009-2010, Viacheslav Tykhanovskyi.
 
 This program is free software, you can redistribute it and/or modify it under
 the terms of the Artistic License version 2.0.
